@@ -5,7 +5,7 @@
 if [ -e ${prefix}/share/kea/scripts/admin-utils.sh ]; then
     . ${prefix}/share/kea/scripts/admin-utils.sh
 else
-    . /usr/obj/ports_mfs/kea-1.1.0/kea-1.1.0/src/bin/admin/admin-utils.sh
+    . /pobj/kea-1.2.0/kea-1.2.0/src/bin/admin/admin-utils.sh
 fi
 
 VERSION=`mysql_version "$@"`
